/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Contributing author: Paul Crozier (SNL)
------------------------------------------------------------------------- */

#include "math.h"
#include "string.h"
#include "fix_spring.h"
#include "atom.h"
#include "update.h"
#include "respa.h"
#include "domain.h"
#include "error.h"
#include "group.h"

#define TETHER 0
#define COUPLE 1

/* ---------------------------------------------------------------------- */

FixSpring::FixSpring(int narg, char **arg) : Fix(narg, arg)
{
  if (narg < 9) error->all("Illegal fix spring command");

  if (strcmp(arg[3],"tether") == 0) {
    if (narg != 9) error->all("Illegal fix spring command");
    styleflag = TETHER;
    k_spring = atof(arg[4]);
    xflag = yflag = zflag = 1;
    if (strcmp(arg[5],"NULL") == 0) xflag = 0;
    else xc = atof(arg[5]);
    if (strcmp(arg[6],"NULL") == 0) yflag = 0;
    else yc = atof(arg[6]);
    if (strcmp(arg[7],"NULL") == 0) zflag = 0;
    else zc = atof(arg[7]);
    r0 = atof(arg[8]);
    if (r0 < 0) error->all("R0 < 0 for fix spring command");

  } else if (strcmp(arg[3],"couple") == 0) {
    if (narg != 11) error->all("Illegal fix spring command");
    styleflag = COUPLE;
    group1 = group->find(arg[4]);
    group2 = group->find(arg[5]);
    if (group1 == -1 || group2 == -1) 
      error->all("Could not find fix spring vector group ID"); 
    group1bit = group->bitmask[group1];
    group2bit = group->bitmask[group2];
    k_spring = atof(arg[6]);
    xflag = yflag = zflag = 1;
    if (strcmp(arg[7],"NULL") == 0) xflag = 0;
    else xc = atof(arg[7]);
    if (strcmp(arg[8],"NULL") == 0) yflag = 0;
    else yc = atof(arg[8]);
    if (strcmp(arg[9],"NULL") == 0) zflag = 0;
    else zc = atof(arg[9]);
    r0 = atof(arg[10]);
    if (r0 < 0) error->all("R0 < 0 for fix spring command");

  } else error->all("Illegal fix spring command");
}

/* ---------------------------------------------------------------------- */

int FixSpring::setmask()
{
  int mask = 0;
  mask |= POST_FORCE;
  mask |= POST_FORCE_RESPA;
  return mask;
}

/* ---------------------------------------------------------------------- */

void FixSpring::init()
{
  // masstotals = total mass of atoms in one or two groups

  int *mask = atom->mask;
  int *type = atom->type;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;

  if (styleflag == TETHER) {
    double rmass = 0.0;
    for (int i = 0; i < nlocal; i++)
      if (mask[i] & groupbit) rmass += mass[type[i]];
    MPI_Allreduce(&rmass,&masstotal,1,MPI_DOUBLE,MPI_SUM,world);
  }

  if (styleflag == COUPLE) {
    double rmass = 0.0;
    for (int i = 0; i < nlocal; i++)
      if (mask[i] & groupbit && mask[i] & group1bit) rmass += mass[type[i]];
    MPI_Allreduce(&rmass,&masstotal1,1,MPI_DOUBLE,MPI_SUM,world);
  
    rmass = 0.0;
    for (int i = 0; i < nlocal; i++)
      if (mask[i] & groupbit && mask[i] & group2bit) rmass += mass[type[i]];
    MPI_Allreduce(&rmass,&masstotal2,1,MPI_DOUBLE,MPI_SUM,world);
  }
  
  if (strcmp(update->integrate_style,"respa") == 0)
    nlevels_respa = ((Respa *) update->integrate)->nlevels;
}

/* ---------------------------------------------------------------------- */

void FixSpring::setup()
{
  if (strcmp(update->integrate_style,"verlet") == 0)
    post_force(1);
  else {
    ((Respa *) update->integrate)->copy_flevel_f(nlevels_respa-1);
    post_force_respa(1,nlevels_respa-1,0);
    ((Respa *) update->integrate)->copy_f_flevel(nlevels_respa-1);
  }
}

/* ---------------------------------------------------------------------- */

void FixSpring::post_force(int vflag)
{
  if (styleflag == TETHER) spring_tether();
  else spring_couple();
}

/* ---------------------------------------------------------------------- */

void FixSpring::spring_tether()
{
  double com[3];
  com_group(com);

  // fx,fy,fz = components of k * (r-r0)
  
  double dx,dy,dz,fx,fy,fz,r,dr;
  
  dx = com[0] - xc;
  dy = com[1] - yc;
  dz = com[2] - zc;
  if (!xflag) dx = 0.0;
  if (!yflag) dy = 0.0;
  if (!zflag) dz = 0.0;
  r = sqrt(dx*dx + dy*dy + dz*dz);
  dr = r - r0;
  fx = k_spring*dx*dr/r;
  fy = k_spring*dy*dr/r;
  fz = k_spring*dz*dr/r;
  
  // apply restoring force to atoms in group
  // f = -k*(r-r0)*mass/masstotal

  double **f = atom->f;
  int *mask = atom->mask;
  int *type = atom->type;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;
  
  double massfrac;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      massfrac = mass[type[i]]/masstotal;
      f[i][0] -= fx*massfrac;
      f[i][1] -= fy*massfrac;
      f[i][2] -= fz*massfrac;
    }
}

/* ---------------------------------------------------------------------- */

void FixSpring::spring_couple()
{
  double com[6];
  com_group2(com);
  
  // fx,fy,fz = components of k * (r-r0)
  
  double dx,dy,dz,fx,fy,fz,r,dr;
  
  dx = com[3] - com[0] - xc;
  dy = com[4] - com[1] - yc;
  dz = com[5] - com[2] - zc;  
  if (!xflag) dx = 0.0;
  if (!yflag) dy = 0.0;
  if (!zflag) dz = 0.0;
  r = sqrt(dx*dx + dy*dy + dz*dz);
  dr = r - r0;
  
  fx = k_spring*dx*dr/r;
  fy = k_spring*dy*dr/r;
  fz = k_spring*dz*dr/r;
  
  // apply restoring force to atoms in group
  // f = -k*(r-r0)*mass/masstotal

  double **f = atom->f;
  int *mask = atom->mask;
  int *type = atom->type;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;
  
  double massfrac;
  for (int i = 0; i < nlocal; i++) {         
    if (mask[i] & groupbit) {
      if (mask[i] & group1bit) {
	massfrac = mass[type[i]]/masstotal1;
	f[i][0] += fx*massfrac;
	f[i][1] += fy*massfrac;
	f[i][2] += fz*massfrac;
      }
      if (mask[i] & group2bit) {
	massfrac = mass[type[i]]/masstotal2;
	f[i][0] -= fx*massfrac;
	f[i][1] -= fy*massfrac;
	f[i][2] -= fz*massfrac;
      }
    }
  }
}

/* ---------------------------------------------------------------------- */

void FixSpring::com_group(double *com_all)
{
  double **x = atom->x;
  int *mask = atom->mask;
  int *type = atom->type;
  int *image = atom->image;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;

  // compute center-of-mass of group
  // use unwrapped coordinates via image
  // else answer is bogus if group straddles a periodic boundary

  double com[3];
  com[0] = com[1] = com[2] = 0.0;
  double xprd = domain->xprd;
  double yprd = domain->yprd;
  double zprd = domain->zprd;
   
  int xbox,ybox,zbox;
  double massone;
  for (int i = 0; i < nlocal; i++)
    if (mask[i] & groupbit) {
      xbox = image[i]%1000 - 500;
      ybox = (image[i]/1000) % 1000 - 500;
      zbox = image[i]/1000000 - 500;
      massone = mass[type[i]];
      com[0] += (x[i][0] + xbox*xprd) * massone;
      com[1] += (x[i][1] + ybox*yprd) * massone;
      com[2] += (x[i][2] + zbox*zprd) * massone;
    }
  
  MPI_Allreduce(com,com_all,3,MPI_DOUBLE,MPI_SUM,world);
  com_all[0] /= masstotal;
  com_all[1] /= masstotal;
  com_all[2] /= masstotal;
}  


/* ---------------------------------------------------------------------- */

void FixSpring::com_group2(double *com_all)
{
  double **x = atom->x;
  int *mask = atom->mask;
  int *type = atom->type;
  int *image = atom->image;
  double *mass = atom->mass;
  int nlocal = atom->nlocal;

  // compute center-of-mass of each group
  // use unwrapped coordinates via image
  // else answer is bogus if group straddles a periodic boundary

  double com[6];
  for (int i = 0; i < 6; i++) com[i] = 0.0;
  double xprd = domain->xprd;
  double yprd = domain->yprd;
  double zprd = domain->zprd;
  
  int xbox,ybox,zbox;
  double massone;
  for (int i = 0; i < nlocal; i++) {
    if (mask[i] & groupbit) {
      if (mask[i] & group1bit) {
	xbox = image[i]%1000 - 500;
	ybox = (image[i]/1000) % 1000 - 500;
	zbox = image[i]/1000000 - 500;
	massone = mass[type[i]];
	com[0] += (x[i][0] + xbox*xprd) * massone;
	com[1] += (x[i][1] + ybox*yprd) * massone;
	com[2] += (x[i][2] + zbox*zprd) * massone;
      }
      if (mask[i] & group2bit) {
	xbox = image[i]%1000 - 500;
	ybox = (image[i]/1000) % 1000 - 500;
	zbox = image[i]/1000000 - 500;
	massone = mass[type[i]];
	com[3] += (x[i][0] + xbox*xprd) * massone;
	com[4] += (x[i][1] + ybox*yprd) * massone;
	com[5] += (x[i][2] + zbox*zprd) * massone;
      }
    }
  }
  
  MPI_Allreduce(com,com_all,6,MPI_DOUBLE,MPI_SUM,world);
  com_all[0] /= masstotal1;
  com_all[1] /= masstotal1;
  com_all[2] /= masstotal1;
  com_all[3] /= masstotal2;
  com_all[4] /= masstotal2;
  com_all[5] /= masstotal2;
}

/* ---------------------------------------------------------------------- */

void FixSpring::post_force_respa(int vflag, int ilevel, int iloop)
{
  if (ilevel == nlevels_respa-1) post_force(vflag);
}
