/* ----------------------------------------------------------------------
   LAMMPS - Large-scale Atomic/Molecular Massively Parallel Simulator
   www.cs.sandia.gov/~sjplimp/lammps.html
   Steve Plimpton, sjplimp@sandia.gov, Sandia National Laboratories

   Copyright (2003) Sandia Corporation.  Under the terms of Contract
   DE-AC04-94AL85000 with Sandia Corporation, the U.S. Government retains
   certain rights in this software.  This software is distributed under 
   the GNU General Public License.

   See the README file in the top-level LAMMPS directory.
------------------------------------------------------------------------- */

#include "math.h"
#include "string.h"
#include "unistd.h"
#include "variable.h"
#include "universe.h"
#include "atom.h"
#include "update.h"
#include "domain.h"
#include "output.h"
#include "thermo.h"
#include "memory.h"
#include "error.h"

#define MAXVAR 26

enum{INDEX,LOOP,EQUAL,WORLD,UNIVERSE};

/* ---------------------------------------------------------------------- */

Variable::Variable()
{
  MPI_Comm_rank(world,&me);

  num = new int[MAXVAR];
  style = new int[MAXVAR];
  index = new int[MAXVAR];
  data = new char**[MAXVAR];

  for (int i = 0; i < MAXVAR; i++) num[i] = 0;
  for (int i = 0; i < MAXVAR; i++) data[i] = NULL;
}

/* ---------------------------------------------------------------------- */

Variable::~Variable()
{
  int i,j;

  // free data and other arrays
  // each variable style has different data values

  for (i = 0; i < MAXVAR; i++) {
    if (num[i] == 0) continue;
    if (style[i] == INDEX) {
      for (j = 0; j < num[i]; j++) delete [] data[i][j];
    } else if (style[i] == LOOP) {
      if (data[i][0]) delete [] data[i][0];
    } else if (style[i] == EQUAL) {
      delete [] data[i][0];
      if (data[i][1]) delete [] data[i][1];
    } else if (style[i] == WORLD) {
      for (j = 0; j < num[i]; j++) delete [] data[i][j];
    } else if (style[i] == UNIVERSE) {
      for (j = 0; j < num[i]; j++) delete [] data[i][j];
    }
    delete [] data[i];
  }
  if (data) delete [] data;

  delete [] num;
  delete [] style;
  delete [] index;
}

/* ----------------------------------------------------------------------
   called by variable command in input script to create INDEX or PROC var
------------------------------------------------------------------------- */

void Variable::set(int narg, char **arg)
{
  if (narg < 3) error->all("Illegal variable command");

  // ivar = which variable, a to z = 0 to 25

  int ivar = arg[0][0] - 'a';
  if (ivar < 0 || ivar >= MAXVAR) 
    error->all("Invalid variable in variable command");

  // setup the variable
  // for all styles except EQUAL, if var already exists, just skip

  // INDEX
  // data = list of values
  // index = 1st value

  if (strcmp(arg[1],"index") == 0) {
    if (num[ivar]) {
      if (style[ivar] != INDEX)
	error->all("Cannot redefine variable as a different style");
      return;
    }
    style[ivar] = INDEX;
    num[ivar] = narg - 2;
    data[ivar] = new char*[num[ivar]];
    copy(num[ivar],&arg[2],data[ivar]);
    index[ivar] = 0;

  // LOOP
  // data = single value, only filled when variable is retrieved
  // num = N
  // index = 1st value

  } else if (strcmp(arg[1],"loop") == 0) {
    if (num[ivar]) {
      if (style[ivar] != LOOP)
	error->all("Cannot redefine variable as a different style");
      return;
    }
    style[ivar] = LOOP;
    num[ivar] = atoi(arg[2]);
    data[ivar] = new char*[1];
    data[ivar][0] = NULL;
    index[ivar] = 0;

  // EQUAL
  // delete pre-existing var if also style EQUAL (allows it to be re-set)
  // data = 2 values, 1st is string to eval, 2nd is filled on retrieval
  // num = 1
  // index = 1st value

  } else if (strcmp(arg[1],"equal") == 0) {
    if (num[ivar]) {
      if (style[ivar] != EQUAL)
	error->all("Cannot redefine variable as a different style");
      delete [] data[ivar][0];
      if (data[ivar][1]) delete [] data[ivar][1];
      free(data[ivar]);
    }
    style[ivar] = EQUAL;
    num[ivar] = 1;
    data[ivar] = new char*[2];
    copy(1,&arg[2],data[ivar]);
    data[ivar][1] = NULL;
    index[ivar] = 0;

  // WORLD
  // data = list of values
  // index = partition this proc is in
  // error check that list size = # of worlds in universe

  } else if (strcmp(arg[1],"world") == 0) {
    if (num[ivar]) {
      if (style[ivar] != WORLD)
	error->all("Cannot redefine variable as a different style");
      return;
    }
    style[ivar] = WORLD;
    num[ivar] = narg - 2;
    if (num[ivar] != universe->nworlds)
      error->all("World variable count doesn't match # of partitions");
    data[ivar] = new char*[num[ivar]];
    copy(num[ivar],&arg[2],data[ivar]);
    index[ivar] = universe->iworld;

  // UNIVERSE
  // data = list of values
  // universe proc 0 creates lock file
  // index = partition this proc is in
  // error check that all other universe variables are same length

  } else if (strcmp(arg[1],"universe") == 0) {
    if (num[ivar]) {
      if (style[ivar] != UNIVERSE)
	error->all("Cannot redefine variable as a different style");
      return;
    }
    if (universe->me == 0) {
      FILE *fp = fopen("tmp.lammps.variable","w");
      fprintf(fp,"%d\n",universe->nworlds);
      fclose(fp);
    }

    style[ivar] = UNIVERSE;
    num[ivar] = narg - 2;
    if (num[ivar] < universe->nworlds)
      error->all("Universe variable count < # of partitions");
    data[ivar] = new char*[num[ivar]];
    copy(num[ivar],&arg[2],data[ivar]);
    index[ivar] = universe->iworld;

    for (int jvar = 0; jvar < MAXVAR; jvar++)
      if (num[jvar] && style[jvar] == UNIVERSE && num[ivar] != num[jvar])
	error->all("All universe variables must have same # of values");

    if (me == 0) {
      if (universe->uscreen)
	fprintf(universe->uscreen,
		"Initial $%c setting: value %d on partition %d\n",
		arg[0][0],index[ivar]+1,universe->iworld);
      if (universe->ulogfile)
	fprintf(universe->ulogfile,
		"Initial $%c setting: value %d on partition %d\n",
		arg[0][0],index[ivar]+1,universe->iworld);
    }

  } else error->all("Illegal variable command");
}

/* ----------------------------------------------------------------------
   single-value INDEX var created by command-line argument
------------------------------------------------------------------------- */

void Variable::set(char *name, char *value)
{
  int ivar = name[0] - 'a';
  if (ivar < 0 || ivar >= MAXVAR) 
    error->all("Invalid variable in command-line argument");

  if (num[ivar]) error->all("Command-line variable already exists");
  style[ivar] = INDEX;
  num[ivar] = 1;
  data[ivar] = new char*[num[ivar]];
  copy(num[ivar],&value,data[ivar]);
  index[ivar] = 0;
}

/* ----------------------------------------------------------------------
   increment variable(s)
   return 0 if OK if successfully incremented
   return 1 if any variable is exhausted and free the variable to allow re-use
------------------------------------------------------------------------- */

int Variable::next(int narg, char **arg)
{
  int ivar;

  if (narg == 0) error->all("Illegal next command");

  // check that variables exist and are all same style

  for (int iarg = 0; iarg < narg; iarg++) {
    ivar = arg[iarg][0] - 'a';
    if (ivar < 0 || ivar >= MAXVAR) 
      error->all("Invalid variable in next command");
    if (num[ivar] == 0) error->all("Invalid variable in next command");
    if (style[ivar] != style[arg[0][0] - 'a'])
      error->all("All variables in next command must be same style");
  }

  // check for invalid styles EQUAL or WORLD

  int onestyle = style[arg[0][0] - 'a'];
  if (onestyle == EQUAL || onestyle == WORLD)
    error->all("Invalid variable style with next command");

  // increment all variables in list
  // set flag = 1 if any variable is exhausted and free it to allow re-use

  int flag = 0;

  if (onestyle == INDEX)
    for (int iarg = 0; iarg < narg; iarg++) {
      ivar = arg[iarg][0] - 'a';
      index[ivar]++;
      if (index[ivar] >= num[ivar]) {
	flag = 1;
	for (int i = 0; i < num[ivar]; i++) delete [] data[ivar][i];
	delete [] data[ivar];
	num[ivar] = 0;
      }
    }
  else if (onestyle == LOOP)
    for (int iarg = 0; iarg < narg; iarg++) {
      ivar = arg[iarg][0] - 'a';
      index[ivar]++;
      if (index[ivar] >= num[ivar]) {
	flag = 1;
	delete [] data[ivar][0];
	delete [] data[ivar];
	num[ivar] = 0;
      }
    }
  else if (onestyle == UNIVERSE) {

    // wait until lock file can be created and owned by proc 0 of this world
    // read next available index and Bcast it within my world

    int nextindex;
    if (me == 0) {
      while (1) {
	if (rename("tmp.lammps.variable","tmp.lammps.variable.lock")) break;
	usleep(100000);
      }
      FILE *fp = fopen("tmp.lammps.variable.lock","r");
      fscanf(fp,"%d",&nextindex);
      fclose(fp);
      fp = fopen("tmp.lammps.variable.lock","w");
      fprintf(fp,"%d\n",nextindex+1);
      fclose(fp);
      rename("tmp.lammps.variable.lock","tmp.lammps.variable");
      if (universe->uscreen)
	fprintf(universe->uscreen,
		"Increment via next: value %d on partition %d\n",
		nextindex+1,universe->iworld);
      if (universe->ulogfile)
	fprintf(universe->ulogfile,
		"Increment via next: value %d on partition %d\n",
		nextindex+1,universe->iworld);
    }
    MPI_Bcast(&nextindex,1,MPI_INT,0,world);

    // set all variables in next list to nextindex

    for (int iarg = 0; iarg < narg; iarg++) {
      ivar = arg[iarg][0] - 'a';
      index[ivar] = nextindex;
      if (index[ivar] >= num[ivar]) {
	for (int i = 0; i < num[ivar]; i++) delete [] data[ivar][i];
	delete [] data[ivar];
	num[ivar] = 0;
	flag = 1;
      }
    }
  }

  return flag;
}

/* ----------------------------------------------------------------------
   return ptr to the data text associated with a variable
   return NULL if no variable or index is bad, caller must respond
------------------------------------------------------------------------- */

char *Variable::retrieve(char tag)
{
  int ivar = tag - 'a';
  if (num[ivar] == 0) return NULL;                // no variable exists
  if (index[ivar] >= num[ivar]) return NULL;      // variable exhausted

  char *str;
  if (style[ivar] == INDEX) {
    str = data[ivar][index[ivar]];
  } else if (style[ivar] == LOOP) {
    char *value = new char[16];
    sprintf(value,"%d",index[ivar]+1);
    int n = strlen(value) + 1;
    if (data[ivar][0]) delete [] data[ivar][0];
    data[ivar][0] = new char[n];
    strcpy(data[ivar][0],value);
    delete [] value;
    str = data[ivar][0];
  } else if (style[ivar] == EQUAL) {
    char *value = evaluate(data[ivar][0]);
    int n = strlen(value) + 1;
    if (data[ivar][1]) delete [] data[ivar][1];
    data[ivar][1] = new char[n];
    strcpy(data[ivar][1],value);
    delete [] value;
    str = data[ivar][1];
  } else if (style[ivar] == WORLD) {
    str = data[ivar][index[ivar]];
  } else if (style[ivar] == UNIVERSE) {
    str = data[ivar][index[ivar]];
  }
  return str;
}

/* ----------------------------------------------------------------------
   copy narg strings from **from to **to 
------------------------------------------------------------------------- */
  
void Variable::copy(int narg, char **from, char **to)
{
  int n;

  for (int i = 0; i < narg; i++) {
    n = strlen(from[i]) + 1;
    to[i] = new char[n];
    strcpy(to[i],from[i]);
  }
}

/* ----------------------------------------------------------------------
   recursive method to evaluate a string
   string can be a number: 0.0, -5.45, etc
   string can be a keyword: lx, ly, lz, vol, etc
   string can be a vector: x[123], y[3], vx[34], etc
     value inside brackets must be global ID from 1 to Natoms
   string can be a function: div(x,y), mult(x,y), add(x,y), neg(x), etc
     each function arg can be a number, keyword, vector, or function
   see lists of valid keywords, vectors, and functions below
   when string is evaluated, put result in a newly allocated string
   return address of result string (will be freed by caller)
------------------------------------------------------------------------- */

char *Variable::evaluate(char *str)
{
  // allocate a new string for the eventual result

  char *result = new char[32];
  double answer;

  // if string is "function", grab one or two args, evaulate args recursively

  if (strchr(str,'(')) {
    if (str[strlen(str)-1] != ')')
      error->all("Cannot evaluate variable equal command");

    char *ptr = strchr(str,'(');
    int n = ptr - str;
    char *func = new char[n+1];
    strncpy(func,str,n);
    func[n] = '\0';

    char *comma = ++ptr;
    int level = 0;
    while (1) {
      if (*comma == '\0')
	error->all("Cannot evaluate variable equal command");
      else if (*comma == ',' && level == 0) break;
      else if (*comma == ')' && level == 0) break;
      else if (*comma == '(') level++;
      else if (*comma == ')') level--;
      comma++;
    }

    char *arg1,*arg2;
    n = comma - ptr;
    arg1 = new char[n+1];
    strncpy(arg1,ptr,n);
    arg1[n] = '\0';

    if (*comma == ',') {
      ptr = comma + 1;
      comma = &str[strlen(str)-1];
      n = comma - ptr;
      arg2 = new char[n+1];
      strncpy(arg2,ptr,n);
      arg2[n] = '\0';
    } else arg2 = NULL;
    
    double value1,value2;
    char *strarg1,*strarg2;
    strarg1 = evaluate(arg1);
    if (arg2) strarg2 = evaluate(arg2);

    // customize by adding function to this list and to if statement
    // add(x,y),sub(x,y),mult(x,y),div(x,y),neg(x),pow(x,y),exp(x),ln(x)

    if (strcmp(func,"add") == 0) {
      if (!arg2) error->all("Cannot evaluate variable equal command");
      value1 = atof(strarg1);
      value2 = atof(strarg2);
      answer = value1 + value2;
    } else if (strcmp(func,"sub") == 0) {
      if (!arg2) error->all("Cannot evaluate variable equal command");
      value1 = atof(strarg1);
      value2 = atof(strarg2);
      answer = value1 - value2;
    } else if (strcmp(func,"mult") == 0) {
      if (!arg2) error->all("Cannot evaluate variable equal command");
      value1 = atof(strarg1);
      value2 = atof(strarg2);
      answer = value1 * value2;
    } else if (strcmp(func,"div") == 0) {
      if (!arg2) error->all("Cannot evaluate variable equal command");
      value1 = atof(strarg1);
      value2 = atof(strarg2);
      if (value2 == 0.0)
	error->all("Cannot evaluate variable equal command");
      answer = value1 / value2;
    } else if (strcmp(func,"neg") == 0) {
      if (arg2) error->all("Cannot evaluate variable equal command");
      value1 = atof(strarg1);
      answer = -value1;
    } else if (strcmp(func,"pow") == 0) {
      if (!arg2) error->all("Cannot evaluate variable equal command");
      value1 = atof(strarg1);
      value2 = atof(strarg2);
      if (value2 == 0.0)
	error->all("Cannot evaluate variable equal command");
      answer = pow(value1,value2);
    } else if (strcmp(func,"exp") == 0) {
      if (arg2) error->all("Cannot evaluate variable equal command");
      value1 = atof(strarg1);
      answer = exp(value1);
    } else if (strcmp(func,"ln") == 0) {
      if (arg2) error->all("Cannot evaluate variable equal command");
      value1 = atof(strarg1);
      if (value1 == 0.0)
	error->all("Cannot evaluate variable equal command");
      answer = log(value1);

    } else error->all("Cannot evaluate variable equal command");

    delete [] func;
    delete [] arg1;
    delete [] strarg1;
    if (arg2) {
      delete [] arg2;
      delete [] strarg2;
    }
    sprintf(result,"%g",answer);

  // if string is "vector", find which proc owns atom, grab vector value

  } else if (strchr(str,'[')) {
    if (str[strlen(str)-1] != ']')
      error->all("Cannot evaluate variable equal command");
    if (atom->map_style == 0)
      error->all("Cannot use vectors in variables unless atom map exists");

    char *ptr = strchr(str,'[');
    int n = ptr - str;
    char *vector = new char[n+1];
    strncpy(vector,str,n);
    vector[n] = '\0';

    char *arg;
    ptr++;
    char *ptr2 = &str[strlen(str)-1];
    n = ptr2 - ptr;
    arg = new char[n+1];
    strncpy(arg,ptr,n);
    arg[n] = '\0';
    int i = atom->map(atoi(arg));

    double mine;
    if (i >= 0 && i < atom->nlocal) {

      // customize by adding vector to this list and to if statement
      // x,y,z,vx,vy,vz,fx,fy,fz

      if (strcmp(vector,"x") == 0) mine = atom->x[i][0];
      else if (strcmp(vector,"y") == 0) mine = atom->x[i][1];
      else if (strcmp(vector,"z") == 0) mine = atom->x[i][2];
      else if (strcmp(vector,"vx") == 0) mine = atom->v[i][0];
      else if (strcmp(vector,"vy") == 0) mine = atom->v[i][1];
      else if (strcmp(vector,"vz") == 0) mine = atom->v[i][2];
      else if (strcmp(vector,"fx") == 0) mine = atom->f[i][0];
      else if (strcmp(vector,"fy") == 0) mine = atom->f[i][1];
      else if (strcmp(vector,"fz") == 0) mine = atom->f[i][2];

      else error->one("Invalid vector in variable equal command");

    } else mine = 0.0;
    MPI_Allreduce(&mine,&answer,1,MPI_DOUBLE,MPI_SUM,world);

    delete [] vector;
    delete [] arg;
    sprintf(result,"%g",answer);

  // if string is "keyword", compute appropriate value via thermo

  } else if (str[0] - 'a' >= 0 && str[0] - 'a' < MAXVAR) {

    // customize by adding keyword to this list and to if statement
    // keywords defined by thermo_style custom are passed to compute_value()

    if (domain->box_exist == 0)
      error->all("Using variable equal keyword before simulation box is defined");

    // process these keywords here, not in thermo
    // they are allowed anytime after simulation box is defined

    if (strcmp(str,"step") == 0)
      answer = update->ntimestep;
    else if (strcmp(str,"atoms") == 0)
      answer = atom->natoms;
    else if (strcmp(str,"lx") == 0)
      answer = domain->xprd;
    else if (strcmp(str,"ly") == 0)
      answer = domain->yprd;
    else if (strcmp(str,"lz") == 0)
      answer = domain->zprd;
    else if (strcmp(str,"vol") == 0)
      answer = domain->xprd * domain->yprd * domain->zprd;

    // process other keywords via thermo
    // error if 1st run has not yet started
    // warning if out-of-sync with thermo, since values can be bogus

    else {
      if (update->first_update == 0)
	error->all("Using variable equal keyword before initial run");
      if (update->ntimestep != output->next_thermo && me == 0)
	error->warning("Using variable equal keyword with non-current thermo");
      int flag = output->thermo->compute_value(str,&answer);
      if (flag) error->all("Invalid keyword in variable equal command");
    }
    
    sprintf(result,"%g",answer);

  // string is a number, just copy to result

  } else strcpy(result,str);

  // return newly allocated string

  return result;
}
